/*
 Camera Module

  - connects to an MQTT server
  - publishes "AWUPWUPWUPWUP" to the topic "Zoidburg"
  - subscribes to the topic CAM_SUB_TOPIC defined below
  - Use the serial moniter to see arrived messages w/ Baud rate 38400
  - Takes picture when TAKE_PIC_ACK is sent to topic CAM_SUB_TOPIC, both defined below
  - Uses TX and RX to communicate with UART camera
    -Attach Green to 13 on DevThing and TX on Camera
    -Attach Purple to 15 on DevThing and RX on Camera
    -Attach Red to Vin
    -Attach Black to GND
  - If DOORBELL_PIN goes HIGH, take a picture
  - If it cannot connect to MQTT server, sets up as an Access Point
    - Waits for Network and Server Parameters to be sent to Access Point
    - Once it's connected to MQTT server, destroy Access Point
   
*/

#include <ESP8266WiFi.h>
#include <PubSubClient.h>


/************************* WiFi Access Point *********************************/

//#define WLAN_SSID       "SDNet"
//#define WLAN_PASS       "CapstoneProject"
String Wlan_Ssid=       "Test";
String Wlan_Pass=       "Test";

/************************* Adafruit.io Setup *********************************/

//#define SERVER_ADDRESS      "10.176.58.5"   // server in 213 SR
#define ESP_SERVER_ADDRESS "192.168.4.1"
#define SERVER_PORT         1883            // standard port
bool connected =            false;
String Server_Address=      "Test";
const char WiFiAPPSK[] =    "LockStars";    // Password when in server mode

/************************* Camera-DevThing Parameters ************************/

#define CAM_SUB_TOPIC "Take_Pic"      // Topic camera is listening to
#define CAM_PUB_TOPIC "LSPic"         // Topic camera publishes to
#define TAKE_PIC_ACK "yes"            // ACK to take pic
#define PIC_STEP_SIZE 64              // Receive PIC_STEP_SIZE bytes of data at a time from camera, must be a multiple of 2 
#define DOORBELL_PIN 4                // Pin to connect to doorbell
#define AP_PIN 12                     // Pin set HIGH if acting as an Access Point

/************************* Functions *****************************************/

// Create an ESP8266 WiFiClient class to connect to the MQTT server.
WiFiClient wf_client; // instantiate wifi client
PubSubClient client(wf_client, ESP_SERVER_ADDRESS); // pass to pubsub

void callback(const MQTT::Publish& pub);
void sendResetCmd();
void sendSnapCmd();
void sendSnapLengthCmd();
void getVersionCmd();
int sendReadPicCmd(int a);
void setSizeCmd();
void camBuffDump();
void Take_And_Send_Pic_Cmd();
void setupWiFiAP();
void stopWiFiAP();


/************************* Setup Console and Camera ****************************/

void setup() {
  
  // Setup Serial Monitor
  Serial.begin(38400);
  delay(2000);
  Serial.println("ESP8266 Camera Module");

  // Set doorbell pin as input
  pinMode(DOORBELL_PIN, INPUT);
  delay(10);

  // Set AP_LED pin to output
  pinMode(AP_PIN, OUTPUT);
  digitalWrite(AP_PIN, LOW);
  
  // Convert Strings to char
  char WLAN_SSID[100];
  char WLAN_PASS[100];
  char SERVER_ADDRESS[100];
  Wlan_Ssid.toCharArray(WLAN_SSID, sizeof(WLAN_SSID));
  Wlan_Pass.toCharArray(WLAN_PASS, sizeof(WLAN_PASS));
  Server_Address.toCharArray(SERVER_ADDRESS, sizeof(SERVER_ADDRESS));
  Serial.print("WLAN_SSID: ");
  Serial.println(WLAN_SSID);
  Serial.print("WLAN_PASS: ");
  Serial.println(WLAN_PASS);
  Serial.print("SERVER_ADDRESS: ");
  Serial.print(SERVER_ADDRESS);

  // Create PubSubClient Instance
  client = PubSubClient(wf_client, SERVER_ADDRESS);
  
  // Connect to WiFi server
  Serial.println(); Serial.println();
  Serial.print("Connecting to ");
  Serial.println(Wlan_Ssid);
  delay(100);

  int attemptNum=0;
  int EndFlag=0;
  int i=0;
  WiFi.begin(WLAN_SSID, WLAN_PASS);
  digitalWrite(AP_PIN, HIGH);
  while ((WiFi.status() != WL_CONNECTED)&&(EndFlag==0)) {
    delay(500);
    Serial.print(".");
    attemptNum++;
    if (attemptNum>20){
      attemptNum=0;
      Serial.println("");
      Serial.println("Could Not Connect");
      EndFlag=1;  
    }    
  }

  // Set client callback if connected
  if (WiFi.status() == WL_CONNECTED) {
    digitalWrite(AP_PIN, LOW);
    client.set_callback(callback);
    Serial.println();
    Serial.println("WiFi connected");
    Serial.println("IP address: "); Serial.println(WiFi.localIP());
  }

  // Reset camera and set the picture size
  Serial.swap();
  camBuffDump(); 
  sendResetCmd();
  delay(2);
  setSizeCmd();
  delay(2);
  Serial.swap();

}

void loop() {

  // If WiFi is connected, run program
  if (WiFi.status() == WL_CONNECTED) {
    if (!client.connected()) {
      if (client.connect("Camera_Module")) {
        Serial.println("Connected to MQTT server");
        client.publish("Zoidburg", "AWUPWUPWUPWUPWUP");
        client.subscribe(CAM_SUB_TOPIC);
      }
    }
    
    /* wait for incoming messages */
    if (client.connected())
      client.loop();

    // If doorbell is pressed, take a picture
    int val=digitalRead(DOORBELL_PIN);
    if (val==HIGH){
      Take_And_Send_Pic_Cmd();
      delay(10);
    }
  }

  // If WiFi is not connected, start as access point and look for WiFi information
  else {
    digitalWrite(AP_PIN, HIGH);
    setupWiFiAP();
    WiFiServer server(80);
    server.begin();
    delay(100);
    WiFiClient esp_client= server.available();
    if (!esp_client){
      return;
    }
    String trash;
    String inString;
    while(!esp_client.available()){
    }
    // Read information sent to access point
    // WLAN_SSID
    trash=esp_client.readStringUntil('/');
    Wlan_Ssid=esp_client.readStringUntil('/');

    // WLAN_PASS
    Wlan_Pass=esp_client.readStringUntil('/');

    // SERVER_ADDRESS
    Server_Address=esp_client.readStringUntil(' ');
    trash=esp_client.readStringUntil('\r');
    esp_client.flush();
    
    stopWiFiAP();
    setup();
  }
}

/************************* Functions *****************************************/

void callback(const MQTT::Publish& pub) {
  // handle message arrived
  Serial.print("Message arrived [");
  Serial.print(pub.topic());
  Serial.print("] ");

  Serial.println(pub.payload_string());
  delay(10);

  // If the message says to take a picture, then take a picture
  if(pub.payload_string()==TAKE_PIC_ACK){
    Take_And_Send_Pic_Cmd();
  }
}// end of callback function

// Reset camera
void sendResetCmd(){
    int EndFlag=0;
    char v0, v1, v2, v3, v4, v5, v6, v7, v8, v9;
  
    Serial.write((byte)0x56);
    Serial.write((byte)0x00);
    Serial.write((byte)0x26);
    Serial.write((byte)0x00);

    while(!EndFlag){
        v9=v8;v8=v7;v7=v6;v6=v5;v5=v4;v4=v3;v3=v2;v2=v1;v1=v0;
        while(!Serial.available()>0);
        v0=Serial.read();
        if ((v4==(byte)0x65)&&(v3==(byte)0x6E)&&(v2==(byte)0x64)
        &&(v1==(byte)0x0D)&&(v0==(byte)0x0A)){
          EndFlag=1; 
        }
    }
}

// Tell camera to take picture
void sendSnapCmd(){
    char v0, v1, v2, v3, v4;

    Serial.write((byte)0x56);
    Serial.write((byte)0x00);
    Serial.write((byte)0x36);
    Serial.write((byte)0x01);
    Serial.write((byte)0x00);

    while(!Serial.available()>0);
    v0 = Serial.read();
    while(!Serial.available()>0);    
    v1 = Serial.read();
    while(!Serial.available()>0);
    v2 = Serial.read();
    while(!Serial.available()>0);    
    v3 = Serial.read();
    while(!Serial.available()>0);    
    v4 = Serial.read();
}

// Tell camera to say how long the picture is
void sendSnapLengthCmd(){

    Serial.write((byte)0x56);
    Serial.write((byte)0x00);
    Serial.write((byte)0x34);
    Serial.write((byte)0x01);
    Serial.write((byte)0x00);
}

// Tell Camera to send it's version
void getVersionCmd(){
    char v1, v2, v3, v4, v5, v6, v7, v8, v9, v10, v11, v12, v13, v14, v15, v16;
    
    Serial.write((byte)0x56);
    Serial.write((byte)0x00);
    Serial.write((byte)0x11);
    Serial.write((byte)0x00);
    
    // retrieve the version
    while(!Serial.available()>0);
    v1 = Serial.read();
    while(!Serial.available()>0);    
    v2 = Serial.read();
    while(!Serial.available()>0);
    v3 = Serial.read();
    while(!Serial.available()>0);    
    v4 = Serial.read();
    while(!Serial.available()>0);    
    v5 = Serial.read();
    while(!Serial.available()>0);    
    v6 = Serial.read();
    while(!Serial.available()>0);    
    v7 = Serial.read();
    while(!Serial.available()>0);    
    v8 = Serial.read();
    while(!Serial.available()>0);    
    v9 = Serial.read();
    while(!Serial.available()>0);    
    v10 = Serial.read();
    while(!Serial.available()>0);    
    v11 = Serial.read();
    while(!Serial.available()>0);    
    v12 = Serial.read();
    while(!Serial.available()>0);    
    v13 = Serial.read();
    while(!Serial.available()>0);    
    v14 = Serial.read();
    while(!Serial.available()>0);    
    v15 = Serial.read();
    while(!Serial.available()>0);    
    v16 = Serial.read();
}

// Tell camera that the picture is no longer needed
void sendStopSnapCmd(){
    int EndFlag=0;
    char v0, v1, v2, v3, v4;
    
    Serial.write((byte)0x56);
    Serial.write((byte)0x00);
    Serial.write((byte)0x36);
    Serial.write((byte)0x01);
    Serial.write((byte)0x03);

    EndFlag=0;
    while(!EndFlag){
        v4=v3; v3=v2; v2=v1; v1=v0;
        while(!Serial.available()>0);
        v0=Serial.read();
        if ((v4==(byte)0x76)&&(v3==(byte)0x00)&&(v2==(byte)0x36)
            &&(v1==(byte)0x00)&&(v0==(byte)0x00)){
          EndFlag=1;
        }
    }
}

// Set picture size to 320*240
void setSizeCmd(){ 
    int EndFlag=0;
    char v0, v1, v2, v3, v4;
    
    Serial.write((byte)0x56);
    Serial.write((byte)0x00);
    Serial.write((byte)0x31);
    Serial.write((byte)0x05);
    Serial.write((byte)0x04);
    Serial.write((byte)0x01);
    Serial.write((byte)0x00);
    Serial.write((byte)0x19);
    Serial.write((byte)0x11);

    EndFlag=0;
    while(!EndFlag){
        v4=v3; v3=v2; v2=v1; v1=v0;
        while(!Serial.available()>0);
        v0=Serial.read();
        if ((v4==(byte)0x76)&&(v3==(byte)0x00)&&(v2==(byte)0x31)
            &&(v1==(byte)0x00)&&(v0==(byte)0x00)){
          EndFlag=1;
        }
    }
}

// Clear buffer
void camBuffDump(){
  char trash;
  while(Serial.available()>0){
      trash=Serial.read();
    }
}

// Tell Camera to send part of the picture
int sendReadPicCmd(int a){
    unsigned char MH = a/0x100;
    unsigned char ML = a%0x100;
    unsigned char LH = PIC_STEP_SIZE&0xFF00;
    unsigned char LL = PIC_STEP_SIZE&0x00FF;
    
    Serial.write((byte)0x56);
    Serial.write((byte)0x00);
    Serial.write((byte)0x32);
    Serial.write((byte)0x0c);
    Serial.write((byte)0x00);
    Serial.write((byte)0x0a);
    Serial.write((byte)0x00);
    Serial.write((byte)0x00);
    Serial.write(MH);
    Serial.write(ML);
    Serial.write((byte)0x00);
    Serial.write((byte)0x00);
    Serial.write(LH);
    Serial.write(LL);
    Serial.write((byte)0x00);
    Serial.write((byte)0x0a);
    
    a = a + PIC_STEP_SIZE;
    return a;
}

void Take_And_Send_Pic_Cmd(){

    char v0, v1, v2, v3, v4, v5, v6, v7, v8, v9;
    int a=0x0000, EndFlag=0;
    Serial.swap();
    
    // Tell camera to take picture
    sendSnapCmd();
    delay(2);
    
    a=0x0000;
    int i=0, j=0, count=0;
    EndFlag=0;
    byte pic[PIC_STEP_SIZE];
    byte pic_temp;

    // Receive picture in steps
    while(!EndFlag){
      i=0;
      j=0;
      count=0;
      delay(2);
      camBuffDump();
      a=sendReadPicCmd(a);
      while(j<PIC_STEP_SIZE&&!EndFlag){
        while(!Serial.available());
        v0=Serial.read();
        i++;
        if((i>5)&&(j<PIC_STEP_SIZE)&&(!EndFlag)){
          pic[j]=v0;
          if((pic[j-1]==0xFF)&&(pic[j]==0xD9)) //Check if the picture is over
            EndFlag=1;
          j++;
          count++;
          }
      }
      // Send part of picture over MQTT in steps
      client.publish(CAM_PUB_TOPIC, pic, PIC_STEP_SIZE);
    }
       
    Serial.swap();
    Serial.println("Picture has been Received");
    delay(100);
    Serial.swap();
    camBuffDump();

    // Tell camera picture is no longer needed
    sendStopSnapCmd();
    delay(2);
    Serial.swap();
}

// Create a WiFi access point
void setupWiFiAP(){
  
  WiFi.mode(WIFI_AP);

  // Do a little work to get a unique name. Append the
  // last two bytes of the MAC (HEX'd) to "Thing-":
  uint8_t mac[WL_MAC_ADDR_LENGTH];
  WiFi.softAPmacAddress(mac);
  String macID = String(mac[WL_MAC_ADDR_LENGTH - 2], HEX) +
                 String(mac[WL_MAC_ADDR_LENGTH - 1], HEX);
  macID.toUpperCase();
  String AP_NameString = "LockStarsCamera " + macID;

  char AP_NameChar[AP_NameString.length() + 1];
  memset(AP_NameChar, 0, AP_NameString.length() + 1);

  for (int i=0; i<AP_NameString.length(); i++)
    AP_NameChar[i] = AP_NameString.charAt(i);

  WiFi.softAP(AP_NameChar, WiFiAPPSK);
}

// Kill WiFi access point and start as a server
void stopWiFiAP(){

  WiFi.mode(WIFI_OFF);
  WiFi.softAPdisconnect(true);
  WiFi.mode(WIFI_STA);
  
}

