/*
 Solenoid Module

  - connects to an MQTT server
  - publishes "AWUPWUPWUPWUP" to the topic "Zoidburg"
  - subscribes to the topic SOL_SUB_TOPIC defined below
  - Use the serial moniter to see arrived messages w/ Baud rate 38400
  - If a float is sent between 0 and MAX_OPEN_LEN seconds, the door will open for that long
  - If DOORBELL_PIN goes HIGH, send a signal to take a picture
  - If it cannot connect to MQTT server, sets up as an Access Point
    - Waits for Network and Server Parameters to be sent to Access Point
    - Once it's connected to MQTT server, destroy Access Point
    
*/

#include <ESP8266WiFi.h>
#include <PubSubClient.h>


/************************* WiFi Access Point *********************************/

//#define WLAN_SSID       "SDNet"
//#define WLAN_PASS       "CapstoneProject"
String Wlan_Ssid=       "Test";
String Wlan_Pass=       "Test";

/************************* Adafruit.io Setup *********************************/

//#define SERVER_ADDRESS      "10.176.58.5"   // server in 213 SR
#define ESP_SERVER_ADDRESS  "192.168.4.1"
#define SERVER_PORT         1883            // standard port
bool connected =            false;
String Server_Address=      "Test";
const char WiFiAPPSK[] =    "LockStars";    // Password when in server mode

/************************* Solenoid-DevThing Parameters ************************/

#define SOL_SUB_TOPIC "Open_Door"     // Topic Solenoid is listening to
#define OPEN_DOOR_ACK "yes"           // ACK to Open Door
#define SOL_OUT_PIN 5                // Output Pin to Control Solenoid
#define MAX_OPEN_LEN 60               // Max amount of time (in seconds) the solenoid can be on
String inString="";                   // Used for input string
#define DOORBELL_PIN 4                // Pin to connect to doorbell
#define CAM_SUB_TOPIC "Take_Pic"      // Topic camera is listening to
#define TAKE_PIC_ACK "yes"            // ACK to take pic

/************************* Functions *****************************************/

// Create an ESP8266 WiFiClient class to connect to the MQTT server.
WiFiClient wf_client; // instantiate wifi client
PubSubClient client(wf_client, ESP_SERVER_ADDRESS); // pass to pubsub

void callback(const MQTT::Publish& pub);
void OpenDoor(int t);
void setupWiFiAP();
void stopWiFiAP();


/************************* Setup Console ******** ****************************/

void setup() {
  // Setup console
  Serial.begin(38400);
  delay(500);
  Serial.println();
  Serial.println();
  Serial.println(F("Modified pubsub client basic code using modified pubsub software"));
  
  // Convert Strings to char
  char WLAN_SSID[100];
  char WLAN_PASS[100];
  char SERVER_ADDRESS[100];
  Wlan_Ssid.toCharArray(WLAN_SSID, sizeof(WLAN_SSID));
  Wlan_Pass.toCharArray(WLAN_PASS, sizeof(WLAN_PASS));
  Server_Address.toCharArray(SERVER_ADDRESS, sizeof(SERVER_ADDRESS));
  Serial.print("WLAN_SSID: ");
  Serial.println(WLAN_SSID);
  Serial.print("WLAN_PASS: ");
  Serial.println(WLAN_PASS);
  Serial.print("SERVER_ADDRESS: ");
  Serial.print(SERVER_ADDRESS);

  // Create PubSubClient Instance
  client = PubSubClient(wf_client, SERVER_ADDRESS);
  
  // Connect to WiFi server
  Serial.println(); Serial.println();
  Serial.print("Connecting to ");
  Serial.println(Wlan_Ssid);
  delay(100);

  int attemptNum=0;
  int EndFlag=0;
  int i=0;
  WiFi.begin(WLAN_SSID, WLAN_PASS);
  while ((WiFi.status() != WL_CONNECTED)&&(EndFlag==0)) {
    delay(500);
    Serial.print(".");
    attemptNum++;
    if (attemptNum>16){
      attemptNum=0;
      Serial.println("");
      Serial.println("Could Not Connect");
      EndFlag=1;  
    }    
  }

  // Set client callback if connected
  if (WiFi.status() == WL_CONNECTED) {
    client.set_callback(callback);
    Serial.println();
    Serial.println("WiFi connected");
    Serial.println("IP address: "); Serial.println(WiFi.localIP());
  }

  // Set solenoid pin as output
  pinMode(SOL_OUT_PIN, OUTPUT);
  digitalWrite(SOL_OUT_PIN, HIGH);

  // Set doorbell pin as input
  pinMode(DOORBELL_PIN, INPUT);
  delay(10);

}

void loop() {

  // If WiFi is connected, run program
  if (WiFi.status() == WL_CONNECTED) {
    if (!client.connected()) {
      if (client.connect("mydevice")) {
        Serial.println("Connected to MQTT server");
        client.publish("Zoidburg", "AWUPWUPWUPWUPWUP");
        client.subscribe(SOL_SUB_TOPIC);
      }
    }

    /* wait for incoming messages */

    if (client.connected())
      client.loop();

    // If doorbell is pressed, take a picture
    int val=digitalRead(DOORBELL_PIN);
    if (val==HIGH){
      client.publish(CAM_SUB_TOPIC, TAKE_PIC_ACK);
      delay(500);
    }
  }

  // If WiFi is not connected, start as access point and look for WiFi information
  else {
    setupWiFiAP();
    WiFiServer server(80);
    server.begin();
    delay(100);
    WiFiClient esp_client= server.available();
    if (!esp_client){
      return;
    }
    String trash;
    String inString;
    while(!esp_client.available()){
    }
    // Read information sent to access point
    // WLAN_SSID
    trash=esp_client.readStringUntil('/');
    Wlan_Ssid=esp_client.readStringUntil('/');

    // WLAN_PASS
    Wlan_Pass=esp_client.readStringUntil('/');

    // SERVER_ADDRESS
    Server_Address=esp_client.readStringUntil(' ');
    trash=esp_client.readStringUntil('\r');
    esp_client.flush();
    
    stopWiFiAP();
    setup();
  }
}

/************************* Functions *****************************************/

void callback(const MQTT::Publish& pub) {
  // handle message arrived
  Serial.print("Message arrived [");
  Serial.print(pub.topic());
  Serial.print("] ");

  inString=pub.payload_string();
  Serial.println(inString);
  delay(10);

  // Convert message to float
  float inFloat=0;
  inFloat=inString.toFloat();
  OpenDoor(inFloat);
  
}// end of callback function

// Opens door for t seconds
void OpenDoor(int t){
  digitalWrite(SOL_OUT_PIN, LOW);
  int i=0;
  while (i<t*1000){
    i=i+100;
    delay(100);
  }
  digitalWrite(SOL_OUT_PIN, HIGH);
}

// Create a WiFi access point
void setupWiFiAP(){
  
  WiFi.mode(WIFI_AP);

  // Do a little work to get a unique name. Append the
  // last two bytes of the MAC (HEX'd) to "Thing-":
  uint8_t mac[WL_MAC_ADDR_LENGTH];
  WiFi.softAPmacAddress(mac);
  String macID = String(mac[WL_MAC_ADDR_LENGTH - 2], HEX) +
                 String(mac[WL_MAC_ADDR_LENGTH - 1], HEX);
  macID.toUpperCase();
  String AP_NameString = "LockStarsLock " + macID;

  char AP_NameChar[AP_NameString.length() + 1];
  memset(AP_NameChar, 0, AP_NameString.length() + 1);

  for (int i=0; i<AP_NameString.length(); i++)
    AP_NameChar[i] = AP_NameString.charAt(i);

  WiFi.softAP(AP_NameChar, WiFiAPPSK);
}

// Kill WiFi access point and start as a server
void stopWiFiAP(){

  WiFi.mode(WIFI_OFF);
  WiFi.softAPdisconnect(true);
  WiFi.mode(WIFI_STA);
  
}
